__author__ = 'giacomov'

from GenericOGIPLike import GenericOGIPLike

__instrument_name = "VERITAS"

class VERITASLike(GenericOGIPLike):

    def __init__(self, name, udara_style_root_file):

        super(VERITASLike, self).__init__(name, )